/* Handling of recursive retrieving.
   Copyright (C) 1995,96  Hrvoje Niksic
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif /* HAVE_STRING_H */
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif /* HAVE_UNISTD_H */
#include <errno.h>
#include <assert.h>
#include <ctype.h>

#include "wget.h"
#include "options.h"
#include "url.h"
#include "recur.h"
#include "utils.h"
#include "retr.h"
#include "http.h"
#include "ftp.h"
#include "mtch.h"
#include "host.h"

extern struct options opt;

extern char *version_string;

/* A slist of downloaded URL-s */
urlpos *urls_downloaded;
slist *urls_html;
slist *tried_hosts;

/* The core of recursive retrieving.  Endless recursion is avoided by
   having all URL-s stored to a linked list of URL-s, which is checked
   before loading any URL. That way no URL can get loaded twice.

   The function also supports specification of maximum recursion depth
   and several more goodies. */
uerr_t
recursive_retrieve(const char *file, const char *this_url,
                   int flags)
{
   static slist *ulist;         /* List of undesirable-to-load
                                   URL-s. */
   static char **forbidden = NULL; /* List of forbidden locations */
   static int depth;            /* Recursion depth */
   
   char *constr, *suf, *filename, *newloc;
   int dt, inl;
   int this_url_ftp;            /* See below the explanation */
   uerr_t err;
   urlinfo *rurl;
   urlpos *url_list, *cur_url;
   char *rfile; /* For robots */
   urlinfo *u;

   if (flags & RCLEANUP)
   {
      free_slist(ulist);
      ulist = NULL;
      free_vec(forbidden);
      forbidden = NULL;
      free_slist(urls_html);
      urls_html = NULL;
      free_urlpos(urls_downloaded);
      urls_downloaded = NULL;
      free_slist(tried_hosts);
      tried_hosts = NULL;
      return RETROK;
   }
   
   assert(this_url != NULL);
   /* If quota was exceeded earlier, bail out. */
   if (opt.quota && (opt.downloaded > opt.quota))
      return QUOTEXC;
   /* Cache the current URL in the list. */
   if (flags & RFIRST_TIME)
   {
      ulist = add_slist(ulist, this_url, 0);
      depth = 1;
      forbidden = NULL;
   }
   else
      ++depth;

   /* Bail out if opt.maxreclevel is exceeded. */
   if ((opt.maxreclevel != 0) && (depth > opt.maxreclevel))
   {
#ifdef DEBUG
      if (opt.debug)
         fprintf(opt.lfile, "Recursion depth %d exceeded max. depth %d.\n",
                 depth, opt.maxreclevel);
#endif
      --depth;
      return RECLEVELEXC;
   }
   assert(file != NULL);

   /* Determine whether this_url is an FTP URL. If it is, it means
      that the retrieval is done through proxy. In that case, FTP
      links will be followed by default and recursion will not be
      turned off when following them. */
   this_url_ftp = (urlproto(this_url) == URLFTP);

   /* Get the URL-s from an HTML file: */
   url_list = get_urls_html(file, this_url, 0);

   /* Decide what to do with each of the URLs. A URL will be loaded if
      it meets several requirements, discussed later. */
   for (cur_url = url_list; cur_url; cur_url = cur_url->next)
   {
      /* If quota was exceeded earlier, bail out. */
      if (opt.quota && (opt.downloaded > opt.quota))
         break;
      /* Parse the URL for convenient use in other functions, as well
         as to get the optimized form. It also checks URL
         integrity. */
      u = newurl();
      if (parseurl(cur_url->url, u) != URLOK)
      {
         DEBUGP("Yuck! A bad URL.\n");
         freeurl(u, 1);
         continue;
      }
      if (u->proto == URLFILE)
      {
         DEBUGP("Nothing to do with file:// around here.\n");
         freeurl(u, 1);
         continue;
      }
      assert(u->url != NULL);
      constr = nstrdup(u->url);
      
      /* Several checkings whether a file is acceptable to load:
         1. check if URL is ftp, and we don't load it
         2. check for relative links (if relative_only is set)
         3. check for domain
         4. check for no-parent
         5. check for excludes
         6. check for suffix
         7. check for same host (if spanhost is unset), with possible
         gethostbyname baggage
         8. check for robots.txt

         Addendum: If the URL is FTP, and it is to be loaded, only the
         domain and suffix settings are "stronger".
         
         Note that .html and (yuck) .htm will get loaded
         regardless of suffix rules (but that is remedied later with
         unlink).

         More time- and memory- consuming tests should be put later on
         the list.  */

      /* inl is set if the URL we are working on (constr) is stored in
         ulist. Its usage is crucial to avoid the constant calls to
         in_slist, which is quite slow. */
      inl = in_slist(ulist, constr);
      
      /* If it is FTP, and FTP is not followed, chuck it out. */
      if (!inl)
         if (u->proto == URLFTP && !opt.follow_ftp && !this_url_ftp)
         {
            DEBUGP("Uh, it is FTP but i'm not in the mood to follow FTP.\n");
            ulist = add_slist(ulist, constr, 0);
            inl = 1;
         }
      /* If it is absolute link and they are not followed, chuck it
         out. */
      if (!inl && u->proto != URLFTP)
         if (opt.relative_only && !(cur_url->flags & URELATIVE))
         {
            DEBUGP("It doesn't really look like a relative link.\n");
            ulist = add_slist(ulist, constr, 0);
            inl = 1;
         }
      /* If its domain is not to be accepted/looked-up, chuck it
         out. */
      if (!inl)
         if (!accept_domain(u))
         {
            DEBUGP("I don't like the smell of that domain.\n");
            ulist = add_slist(ulist, constr, 0);
            inl = 1;
         }
      /* Check for parent directory. */
      if (!inl && opt.no_parent)
      {
         urlinfo *ut = newurl();
         if (parseurl(this_url, ut) != URLOK)
            DEBUGP("Double yuck!  The *parent* URL is broken.\n");
         else
         {
            int i;

            /* We check whether the current directory (u->dir) is a
               subset of parent directory (ut->dir).  E.g. if ut->dir
               is `/something', ut->dir must begin with `/something'
               to avoid damnation.  */
            for (i = 0;
                 u->dir[i] && ut->dir[i] && (u->dir[i] == ut->dir[i]);
                 i++)
               ;
            if (ut->dir[i])
            {
               DEBUGP("Trying to escape parental guidance with no_parent on.\n");
               ulist = add_slist(ulist, constr, 0);
               inl = 1;
            }
         }
         freeurl(ut, 1);
      }
      /* If the file does not match the acceptance list, or is on the
         rejection list, chuck it out.  The same goes for the
         directory exclude-list.  */
      if (!inl && opt.excludes)
      {
         /* We need an absolute directory... */
         char *tmpdir;
         if (*u->dir != '/')
         {
            tmpdir = nmalloc(strlen(u->dir) + 2);
            *tmpdir = '/';
            strcpy(tmpdir + 1, u->dir);
         }
         else
            tmpdir = nstrdup(u->dir);
         if (in_acclist((const char **)opt.excludes, tmpdir, 0))
         {
#ifdef DEBUG
            if (opt.debug)
               fprintf(opt.lfile, "%s (%s) is excluded.\n", constr, tmpdir);
#endif
            ulist = add_slist(ulist, constr, 0);
            inl = 1;
         }
      }
      if (!inl)
      {
         if (!(((suf = suffix(constr)) != NULL)
               && (!strcmp(suf, "html") || !strcmp(suf, "htm"))))
            if (!acceptable(u->file))
            {
#ifdef DEBUG
               if (opt.debug)
                  fprintf(opt.lfile,
                          "%s (%s) does not match acc/rej rules.\n",
                          constr, u->file);
#endif
               ulist = add_slist(ulist, constr, 0);
               inl = 1;
            }
         if (suf)
            free(suf);
      }
      /* Optimize the URL (which includes possible DNS lookup) only
         after all other possibilities have been exhausted. */
      if (!inl)
      {
         if (!opt.simple_check)
            opt_url(u);
         free(constr);
         constr = nstrdup(u->url);
         inl = in_slist(ulist, constr);
         if (!inl && u->proto != URLFTP)
            if (!opt.spanhost && this_url && !same_host(this_url, constr))
            {
               DEBUGP("This is not the same hostname as the parent's.\n");
               ulist = add_slist(ulist, constr, 0);
               inl = 1;
            }
      }
      /* What about robots.txt? */
      if (!inl && opt.use_robots && u->proto == URLHTTP && this_url)
      {
         if (!tried_robots(this_url))
         {
            err = retrieve_robots(this_url, ROBOTS_FILENAME);
            if (err == ROBOTSOK)
            {
               rurl = robots_url(this_url, ROBOTS_FILENAME);
               rfile = url_filename(rurl);
               forbidden = parse_robots(rfile);
               freeurl(rurl, 1);
               free(rfile);
            }
            else
               forbidden = NULL;
         }
         
         if (!robots_match(u, forbidden))
         {
#ifdef DEBUG
            if (opt.debug)
               fprintf(opt.lfile, "Stuffing %s because %s forbids it.\n",
                       this_url, ROBOTS_FILENAME);
#endif
            ulist = add_slist(ulist, constr, 0);
            inl = 1;
         }
      }

      filename = NULL;
      /* If it wasn't chucked out, do something with it. */
      if (!inl)
      {
         DEBUGP("I've decided to load it -> ");
         /* Add it to the list of already-loaded URL-s. */
         ulist = add_slist(ulist, constr, 0);
         /* Automatically followed FTPs will *not* be downloaded
            recursively. */
         if (u->proto == URLFTP && !this_url_ftp)
         {
            /* Don't you adore side-effects? */
            opt.recursive = 0;
         }
         /* Reset its type. */
         dt = 0;
         /* Retrieve it. */
         retrieve_url(constr, &filename, &newloc, this_url, &dt);
         if (u->proto == URLFTP && !this_url_ftp)
         {
            /* Restore... */
            opt.recursive = 1;
         }
         if (newloc)
         {
            free(constr);
            constr = newloc;
         }
         /* In case of convert_links: If there was no error, add it to
            the list of downloaded URLs.  We might need it for
            conversion.  */
         if (opt.convert_links && filename)
         {
            if (dt & RETROKF)
            {
               urls_downloaded = add_url(urls_downloaded, constr, filename);
               /* If the URL is HTML, note it. */
               if (dt & TEXTHTML)
                  urls_html = add_slist(urls_html, filename, NOSORT);
            }
         }
         /* If it is not an error, and it is of type text/html, parse
            it recursively. */
         if ((dt & TEXTHTML) && (dt & RETROKF))
            recursive_retrieve(filename, constr, 0);
#ifdef DEBUG
         else if (opt.debug)
            fprintf(opt.lfile, "%s is not text/html so we don't chase.\n",
                    filename ? filename: "(null)");
#endif
         /* If an suffix-rejected file was loaded only because it was HTML,
            undo the error now */
         if (!acceptable(constr) && filename)
         {
            if (opt.verbose)
               fprintf(opt.lfile, "Killing %s since it should be rejected.\n",
                       filename);
            if (unlink(filename) && !opt.quiet)
               fprintf(opt.lfile, "unlink: %s\n", mystrerror(errno));
            dt &= ~RETROKF;
         }
         /* If everything was OK, and links are to be converted, let's
            store the local filename. */
         if (opt.convert_links && (dt & RETROKF) && (filename != NULL))
         {
            cur_url->flags |= UCONVERT;
            cur_url->local_name = nstrdup(filename);
         }
      }
#ifdef DEBUG
      else if (opt.debug)
         fprintf(opt.lfile, "%s already in list, so we don't load.\n", constr);
#endif
      /* Free filename and constr. */
      if (filename)
         free(filename);
      if (constr)
         free(constr);
      freeurl(u, 1);
      /* Increment the pbuf for the appropriate size. */
   }

   if (opt.convert_links)
      convert_links(file, url_list);

   /* Free the linked list of URL-s. */
   free_urlpos(url_list);
   
   /* Decrement the recursion depth. */
   --depth;
   if (opt.quota && (opt.downloaded > opt.quota))
      return QUOTEXC;
   else
      return RETROK;
}

/* If a URL was referenced in more than one file, it convert_links
   will replace it only within the first sweep.  This function
   remedies this by going through all the downloaded HTML files in
   urls_html and checking whether each URL was downloaded (if it is in
   urls_downloaded list).  If it was, it is flagged as UCONVERT and is
   assigned a local name.  */
void
convert_all_links(void)
{
   uerr_t res;
   urlpos *l1, *l2, *urls;
   urlinfo *u;
   slist *html;
   int dl;

   for (html = urls_html; html; html = html->next)
   {
#ifdef DEBUG
      if (opt.debug)
         fprintf(opt.lfile, "Rescanning %s\n", html->string);
#endif
      /* Parse the HTML file... */
      urls = get_urls_html(html->string, NULL, 1);
      if (!urls)
         continue;
      for (l1 = urls; l1; l1 = l1->next)
      {
         /* The URL must be in canonical form to be compared. */
         u = newurl();
         res = parseurl(l1->url, u);
         if (res != URLOK)
         {
            freeurl(u, 1);
            continue;
         }
         /* We decide whether a URL is to be converted according to
            whether it was downloaded.  */
         dl = 0;
         for (l2 = urls_downloaded; l2; l2 = l2->next)
            if (!strcmp(l2->url, u->url))
            {
               dl = 1;
#ifdef DEBUG
               if (opt.debug)
                  fprintf(opt.lfile,
                          "%s flagged for conversion, local %s\n",
                          l2->url, l2->local_name);
#endif
               break;
            }
         if (dl)
         {
            l1->flags |= UCONVERT;
            l1->local_name = nstrdup(l2->local_name);
         }
         freeurl(u, 1);
      }
      /* Convert the links in the file. */
      convert_links(html->string, urls);
      /* Free the data. */
      free_urlpos(urls);
   }
}


/*******************
 * Robots support:
 *******************/

/* Construct the robots URL. */
urlinfo *
robots_url(const char *url, const char *robots_filename)
{
   urlinfo *u = newurl();
   uerr_t err;

   err = parseurl(url, u);
   assert(err == URLOK && u->proto == URLHTTP);
   free(u->file);
   free(u->dir);
   free(u->url);
   u->dir = nstrdup("");
   u->file = nstrdup(robots_filename);
   u->url = str_url(u, 0);
   return u;
}

/* Retrieves the robots_filename from the root server directory, if
   possible. Returns ROBOTSOK if robots were retrieved OK, and
   NOROBOTS if robots could not be retrieved for any reason. */
uerr_t
retrieve_robots(const char *url, const char *robots_filename)
{
   int dt;
   uerr_t err;
   urlinfo *u;

   u = robots_url(url, robots_filename);
   err = retrieve_url(u->url, NULL, NULL, NULL, &dt);
   freeurl(u, 1);
   if (err == RETROK)
      return ROBOTSOK;
   else
      return NOROBOTS;
}

/* Parses the robots_filename returning the disallowed filenames in a
   malloc-ed vector of character pointers.

   It should be fully compliant with the syntax as described in the
   file norobots.txt, adopted by the robots mailing list
   (robots@webcrawler.com). */
char **
parse_robots(const char *robots_filename)
{
   FILE *fp;
   char **entries;
   char *line, *cmd, *str, *p;
   char *base_version, *version;
   int len, num, i;
   int wgetmatched;           /* Is the part meant for Wget? */

   entries = NULL;

   num = 0;
   fp = fopen(robots_filename, "r");
   if (!fp)
      return NULL;
   
   /* Kill version number. */
   base_version = nstrdup(version_string);
   version = nstrdup(version_string);
   for (p = version; *p; p++)
      *p = tolower(*p);
   for (p = base_version; *p && *p != '/'; p++)
      *p = tolower(*p);
   *p = '\0';

   /* Setting this to 1 means that Wget considers itself under
      restrictions by default, even if the User-Agent field is not
      present. However, if it finds the user-agent set to something
      other than Wget, the rest will be ignored (up to the following
      User-Agent field). Thus you may have something like:
      
      Disallow: 1
      Disallow: 2
      User-Agent: stupid-robot
      Disallow: 3
      Disallow: 4
      User-Agent: Wget*
      Disallow: 5
      Disallow: 6
      User-Agent: *
      Disallow: 7

      In this case the 1, 2, 5, 6 and 7 disallow lines will be
      stored. */
   wgetmatched = 1;
   while ((line = read_whole_line(fp)))
   {
      len = strlen(line);
      /* Destroy <CR> if there is one. */
      if (len && line[len - 1] == '\r')
         line[len - 1] = '\0';
      /* According to specifications, optional space may be at the
         end... */
#ifdef DEBUG
      if (opt.debug)
         fprintf(opt.lfile, "Line: %s\n", line);
#endif
      /* Skip spaces. */
      for (cmd = line; *cmd && isspace(*cmd); cmd++);
      if (!*cmd)
      {
         free(line);
         DEBUGP("(chucked out)\n");
         continue;
      }
      /* Look for ':'. */
      for (str = cmd; *str && *str != ':'; str++);
      if (!*str)
      {
         free(line);
         DEBUGP("(chucked out)\n");
         continue;
      }
      /* Zero-terminate the command. */
      *str++ = '\0';
      /* Look for the string beginning... */
      for (; *str && isspace(*str); str++);
      /* Look for comments and kill them off. */
      for (p = str; *p && !isspace(*p); p++);
      if (isspace(*p) && *(p + 1) == '#')
         *p = '\0';
      if (!strcasecmp(cmd, "User-agent"))
      {
         int match = 0;
         /* Lowercase the agent string. */
         for (p = str; *p; p++)
            *p = tolower(*p);
         /* If the string is '*', it matches. */
         if (*str == '*' && !*(str + 1))
            match = 1;
         else
         {
            /* If the string contains wildcards, we'll run it through
               fnmatch(). */
            if (has_wildcards(str))
            {
               /* If the string contains '/', compare with the full
                  version.  Else, compare it to base_version.  */
               if (strchr(str, '/'))
                  match = !fnmatch(str, version, 0);
               else
                  match = !fnmatch(str, base_version, 0);
            }
            else                /* Substring search */
            {
               if (strstr(version, str))
                  match = 1;
               else
                  match = 0;
            }
         }
         /* If Wget is not matched, skip all the entries up to the
            next User-agent field. */
         wgetmatched = match;
      }
      else if (!wgetmatched)
      {
         free(line);
         DEBUGP("(chucking out since it is not applicable for Wget)\n");
         continue;
      }
      else if (!strcasecmp(cmd, "Disallow"))
      {
         /* If "Disallow" is empty, the robot is welcome.  */
         if (!*str)
         {
            free_vec(entries);
            entries = (char **)nmalloc(sizeof(char *));
            *entries = NULL;
         }
         else
         {
            entries = (char **)nrealloc(entries, (num + 2)* sizeof(char *));
            entries[num] = nstrdup(str);
            entries[++num] = NULL;
            /* Strip trailing spaces, according to specifications. */
            for (i = strlen(str); i >= 0 && isspace(str[i]); i--)
               if (isspace(str[i]))
                  str[i] = '\0';
         }
      }
      else
      {
         /* unknown command */
         DEBUGP("(chucked out)\n");
      }
      free(line);
   }
   fclose(fp);
   free(base_version);
   free(version);
   return entries;
}

/* The function maintains a small slist of hosts from which robots
   have been taken. It parses the URL, and if the robots have not been
   tried on that host, it adds them to the slist and returns 0,
   otherwise it just returns 1. */
int
tried_robots(const char *url)
{
   uerr_t err;
   urlinfo *u;

   u = newurl();

   DEBUGP("Checking whether to load robots... ");
   err = parseurl(url, u);
   assert(err == URLOK && u->proto == URLHTTP);
   if (!in_slist(tried_hosts, u->host))
   {
      DEBUGP("yes.\n");
      tried_hosts = add_slist(tried_hosts, u->host, 0);
      freeurl(u, 1);
      return 0;
   }
   else
   {
      DEBUGP("no.\n");
      freeurl(u, 1);
      return 1;
   }
}

/* May the URL url be loaded according to disallowing rules stored in
   forbidden? */
int
robots_match(urlinfo *u, char **forbidden)
{
   int l;

   if (!forbidden)
      return 1;
#ifdef DEBUG
   if (opt.debug)
      fprintf(opt.lfile, "Matching %s against: ", u->path);
#endif
   for (; *forbidden; forbidden++)
   {
#ifdef DEBUG
      if (opt.debug)
         fprintf(opt.lfile, "%s ", *forbidden);
#endif
      l = strlen(*forbidden);
      /* If dir is forbidden, we may not load the file. */
      if (strncmp(u->path, *forbidden, l) == 0)
      {
         DEBUGP("matched.\n");
         return 0; /* Matches, i.e. does not load... */
      }
   }
   DEBUGP("not matched.\n");
   return 1;
}

