/* Various functions of utilitarian nature.
   Copyright (C) 1995,96  Hrvoje Niksic
   
   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.
   
   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.
   
   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.  */


#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif /* HAVE_CONFIG_H */

#include <stdio.h>
#include <stdlib.h>
#ifdef HAVE_STRING_H
#  include <string.h>
#else
#  include <strings.h>
#endif /* HAVE_STRING_H */
#include <ctype.h>
#include <pwd.h>
#include <sys/types.h>
#ifdef TIME_WITH_SYS_TIME
#  include <sys/time.h>
#  include <time.h>
#else /* not TIME_WITH_SYS_TIME_H */
#ifdef HAVE_SYS_TIME_H
#  include <sys/time.h>
#else /* not HAVE_SYS_TIME_H */
#  include <time.h>
#endif /* HAVE_SYS_TIME_H */
#endif /* TIME_WITH_SYS_TIME_H */
#ifdef HAVE_UNISTD_H
#  include <unistd.h>
#endif

#include "utils.h"
#include "options.h"

extern struct options opt;

long internal_secs, internal_msecs;


/* nmalloc, nrealloc and nstrdup exit the program if there is not
   enough memory. nstrdup also implements strdup on systems that
   do not have it. */
void *
nmalloc(size_t size)
{
   void *res;

   res = malloc(size);
   if (!res)
      memfatal("malloc");
   return res;
}

void *
nrealloc(void *obj, size_t size)
{
   void *res;

   /* Not all Unixes have the feature of realloc() that calling it
      with a NULL-pointer is the same as malloc(), but it is easy to
      simulate. */
   if (obj)
      res = realloc(obj, size);
   else
      res = malloc(size);
   if (!res)
      memfatal("realloc");
   return res;
}

char *
nstrdup(const char *s)
{
   char *s1;

#ifndef HAVE_STRDUP
   int l;
   
   l = strlen(s);
   s1 = malloc(l + 1);
   if (!s1)
      memfatal("strdup");
   memcpy(s1, s, l + 1);
   return s1;
#else
   s1 = strdup(s);
   if (!s1)
      memfatal("strdup");
   return s1;
#endif
}

/* Croak the fatal memory error and bail out with non-zero exit
   status. */
void
memfatal(const char *s)
{
   fprintf(opt.lfile, "%s: Not enough memory.\n", s);
   exit(1);
}

/* Copy the string formed by two pointers (one on the beginning, other
   on the char after the last char) to a new, malloc-ed
   location. 0-terminate it. */
char *
strdupdelim(const char *beg, const char *end)
{
   char *res;
   
   res = (char *)nmalloc(end - beg + 1);
   memcpy(res, beg, end - beg);
   res[end - beg] = '\0';
   return res;
}

/* Free the pointers in a NULL-terminated vector of pointers, then
   free the pointer itself. */
void
free_vec(char **vec)
{
   int i;

   if (!vec)
      return;
   for (i = 0; vec[i]; i++)
      free(vec[i]);
   free(vec);
}

/* A cuserid() immitation using getpwuid(), to avoid hassling with
   utmp.  Besides, not all systems have cuerid(). */
char *
mycuserid(char *where)
{
   struct passwd *pwd;

   if (!(pwd = getpwuid(getuid())) || !pwd->pw_name)
      return NULL;
   if (where)
   {
      strcpy(where, pwd->pw_name);
      return where;
   }
   else
      return pwd->pw_name;
}

/* The function reads a whole line. It reads the line realloc-ing the
   storage exponentially, doubling the storage after each overflow to
   minimize the number of calls to realloc().

   It is not an exemplary of correctness, since it kills off the
   newline (and no, there is no way to know if there was a newline at
   EOF). */
char *
read_whole_line(FILE *fp)
{
   char *line;
   int i, bufsize, c;

   i = 0;
   bufsize = DYNAMIC_LINE_BUFFER;
   line = nmalloc(bufsize);
   /* Construct the line. */
   while ((c = getc(fp)) != EOF && c != '\n')
   {
      if (i > bufsize - 1)
         line = (char *)nrealloc(line, (bufsize <<= 1));
      line[i++] = c;
   }
   if (c == EOF && !i)
   {
      free(line);
      return NULL;
   }
   /* Check for overflow at zero-termination (no need to double the
      buffer in this case. */
   if (i == bufsize)
      line = (char *)nrealloc(line, i + 1);
   line[i] = '\0';
   return line;
}

/* A set of simple-minded routines to store and search for strings in
   a linked list. You may add a string to the slist, and peek whether
   it's still in there at any time later. */

/* Add an element to the list.  If flags is NOSORT, the list will not
   be sorted.  */
slist *
add_slist(slist *l, const char *s, int flags)
{
   slist *t, *old, *beg;
   int cmp;

   if (flags & NOSORT)
   {
      if (!l)
      {
         t = (slist *)nmalloc(sizeof(slist));
         t->string = nstrdup(s);
         t->next = NULL;
         return t;
      }
      beg = l;
      /* Find the last element. */
      while (l->next)
         l = l->next;
      t = (slist *)nmalloc(sizeof(slist));
      l->next = t;
      t->string = nstrdup(s);
      t->next = NULL;
      return beg;
   }
   /* Empty list or changing the first element. */
   if (!l || (cmp = strcmp(l->string, s)) > 0)
   {
      t = (slist *)nmalloc(sizeof(slist));
      t->string = nstrdup(s);
      t->next = l;
      return t;
   }

   beg = l;
   if (cmp == 0)
      return beg;

   /* Second two one-before-the-last element. */
   while (l->next)
   {
      old = l;
      l = l->next;
      cmp = strcmp(s, l->string);
      if (cmp == 0)             /* No repeating in the list. */
         return beg;
      else if (cmp > 0)
         continue;
      /* If the next list element is greater than s, put s between the
         current and the next list element. */
      t = (slist *)nmalloc(sizeof(slist));
      old->next = t;
      t->next = l;
      t->string = nstrdup(s);
      return beg;
   }
   t = (slist *)nmalloc(sizeof(slist));
   t->string = nstrdup(s);
   /* Insert the new element after the last element. */
   l->next = t;
   t->next = NULL;
   return beg;
}   

/* Is there a specific entry in the list? */
int
in_slist(slist *l, const char *s)
{
   int cmp;

   while (l)
   {
      cmp = strcmp(l->string, s);
      if (cmp == 0)
         return 1;
      else if (cmp > 0)         /* The list is ordered! */
         return 0;
      l = l->next;
   }
   return 0;
}

/* Free the whole slist. */
void
free_slist(slist *l)
{
   slist *n;

   while (l)
   {
      n = l->next;
      free(l->string);
      free(l);
      l = n;
   }
}

/* Parse a string containing comma-separated elements, and return
   a vector of char pointers with the elements. The comma-separated
   list is space-insensitive. */
char **
sepstring(const char *s)
{
   char *tok, **res, *p;
   int i, j, l;
   
   if (!s || !*s)
      return NULL;
   /* Strip spaces from s. */
   l = strlen(s);
   p = (char *)nmalloc(l + 1);
   for (i = j = 0; i < l; i++)
      if (!isspace(s[i]))
         p[j++] = s[i];
   p[j] = '\0';
   if (!strchr(p, ','))
   {
      res = (char **)nmalloc(2 * sizeof(char *));
      res[0] = nstrdup(p);
      res[1] = NULL;
      free(p);
      return res;
   }
   tok = strtok(p, ",");
   res = (char **)nmalloc(2 * sizeof(char *));
   res[0] = nstrdup(tok);
   res[1] = NULL;
   for (i = 1; (tok = strtok(NULL, ",")); i++)
   {
      res = (char **)nrealloc(res, (i + 2) * sizeof(char *));
      res[i] = nstrdup(tok);
      res[i + 1] = NULL;
   }
   free(p);
   return res;
}


/* Legible -- return a static pointer to the legibly printed long. */
char *
legible(long l)
{
   static char buf[20];
   char inbuf[20];
   int i, i1, mod;
   char *ptr, *in;

   /* Fill the buffer. */
   sprintf(inbuf, "%ld", l);
   /* Reset the pointers. */
   ptr = buf;
   in = inbuf;
   /* If the number is negative, shift the pointers. */
   if (*in == '-')
   {
      *ptr++ = '-';
      ++in;
   }
   /* How many digits before the first separator? */
   mod = strlen(in) % 3;
   /* Insert them. */
   for (i = 0; i < mod; i++)
      *ptr++ = in[i];
   /* Now insert the rest of them, putting separator before every
      third digit. */
   for (i1 = i, i = 0; in[i1]; i++, i1++)
   {
      if (i % 3 == 0 && i1 != 0)
         *ptr++ = LEGIBLE_SEPARATOR;
      *ptr++ = in[i1];
   }
   /* Zero-terminate the string. */
   *ptr = '\0';
   return buf;
}

/* How many digits in a (long) integer? */
int
numdigit(long a)
{
   int res;

   for (res = 1; a /= 10; res++);
   return res;
}

/* Print the integer to the string buffer.  The digits are first
   written in reverse order (the least significant digit first), and
   are then reversed.  */
void
prnum(char *where, int num)
{
   char *p = where;
   int i = 0, l;
   char c;

   /* Print the digits to the string. */
   do
   {
      *p++ = num % 10 + '0';
      num /= 10;
   } while (num);
   /* And reverse them. */
   l = p - where - 1;
   for (i = l/2; i >= 0; i--)
   {
      c = where[i];
      where[i] = where[l - i];
      where[l - i] = c;
   }
   where[l + 1] = '\0';
}

/* A function to reset the internal timer. */
void
reset_timer(void)
{
#ifdef HAVE_GETTIMEOFDAY
   struct timeval t;
   gettimeofday(&t, NULL);
   internal_secs = t.tv_sec;
   internal_msecs = t.tv_usec / 1000;
#else
   internal_secs = time(NULL);
   internal_msecs = 0;
#endif
}

/* The time elapsed from the last call to reset_timer, in msecs. */
long
elapsed_time(void)
{
#ifdef HAVE_GETTIMEOFDAY
   struct timeval t;
   gettimeofday(&t, NULL);
   return ((t.tv_sec - internal_secs) * 1000
           + (t.tv_usec / 1000 - internal_msecs));
#else
   return (long)time(NULL) - internal_secs;
#endif
}

/* Print out the appropriate download rate.  Appropriate means that if
   rate is > 1024 bytes per second, kilobytes are used, and if rate >
   1024 * 1024 bps, megabytes are used. */
char *
rate(long bytes, long msecs)
{
   static char res[15];
   double dlrate;

   if (!msecs)
      ++msecs;
   dlrate = (double)1000 * bytes / msecs;
   if (dlrate < 1024.0)
      sprintf(res, "%.2fb/s", dlrate);
   else if (dlrate < 1024.0 * 1024.0)
      sprintf(res, "%.2fK/s", dlrate / 1024.0);
   else
      sprintf(res, "%.2fM/s", dlrate / (1024.0 * 1024.0));
   return res;
}

